/*
** Copyright (c) 2002  Hughes Technologies Pty Ltd.  All rights
** reserved.  
**
** Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: main.c,v 1.58 2012/06/04 05:40:02 bambi Exp $
**
*/

/* #define DEBUG_BSD_MALLOC */
/* #define DEBUG_MEMTRACE */

/*
** Module	: main
** Purpose	: Normal single server backend
** Exports	: 
** Depends Upon	: 
*/



/**************************************************************************
** STANDARD INCLUDES
**************************************************************************/

#include <common/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif


/**************************************************************************
** MODULE SPECIFIC INCLUDES
**************************************************************************/

#include <fcntl.h>
#include <limits.h>
#include <locale.h>
#include <errno.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>

#ifdef HAVE_SETRLIMIT
#  include <sys/resource.h>
#endif

#if HAVE_SYS_MMAN_H
#  include <sys/mman.h>
#endif

#include <common/msql_defs.h>
#include <common/debug/debug.h>
#include <common/portability.h>
#include <common/config/config.h>
#include <msqld/index/index.h>
#include <msqld/includes/msqld.h>
#include <msqld/includes/errmsg.h>
#include <msqld/main/main.h>
#include <msqld/main/version.h>
#include <msqld/main/process.h>
#include <msqld/main/table.h>
#include <msqld/main/net.h>
#include <msqld/main/acl.h>
#include <msqld/main/cache.h>
#include <msqld/main/util.h>
#include <msqld/main/parse.h>
#include <msqld/main/tcp.h>
#include <msqld/main/memory.h>
#include <msqld/lock/lock.h>

#ifdef DEBUG_MEMTRACE
#  include </usr/local/include/memtrace.h>
#endif

/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/


/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/


/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/


extern  u_char  *yytext,
		*yyprev;
extern	int	yylineno;
extern	int	yydebug;

extern	char	errMsg[];
extern	char	*packet;

int	curSock,
	numCons = 0,
	maxCons = 0,
	conCount = 0,
	numKids = 0,
	eintrCount = 0,
	startTime,
	explainOnly;

u_int	serverStartTime = 0,
	serverNumCons = 0,
	serverNumQueries = 0;

char	confFile[MSQL_PATH_LEN];

static  char globalQueryBuf[MSQL_GLOBAL_BUF_LEN + 1];


msqld	*globalServer;

char	PROGNAME[] = "msqld",
	BLANK_ARGV[] = "                                                  ";


static char *comTable[] = {
        "???", "Quit", "Init DB", "Query", "DB List", "Table List",
        "Field List", "Create DB", "Drop DB", "Reload ACL",
        "Shutdown", "Index List", "Stats", "Seq Info", "Move DB",
	"Copy DB", "???" };


#if defined(_OS_WIN32)
RETSIGTYPE sigTrap(int sig);
RETSIGTYPE puntServer(int sig);
RETSIGTYPE puntClient(int sig);
#endif




/****************************************************************************
** 	_initServer
**
**	Purpose	: 
**	Args	: 
**	Returns	: 
**	Notes	: 
*/

void initServer(server)
	msqld	*server;
{
	int	tcpPort,
		opt;
	struct	sockaddr_un	unixAddr;
	struct	sockaddr_in	ipAddr;

	cacheSetupTableCache();

        /*
        ** Create an IP socket
        */
        if (configGetIntEntry("system", "remote_access"))
        {
                tcpPort = configGetIntEntry("general", "tcp_port");
                msqlDebug1(MOD_GENERAL,"IP Socket is %d\n",tcpPort);
                server->ipSock = socket(AF_INET, SOCK_STREAM, 0);
                if (server->ipSock < 0)
                {
                        perror("Can't start server : IP Socket ");
			exit(1);
                }
#               ifdef SO_REUSEADDR
                opt = 1;
                setsockopt(server->ipSock, SOL_SOCKET, SO_REUSEADDR, 
			(char *)&opt, sizeof(int));  
#               endif
         
                bzero(&ipAddr, sizeof(ipAddr));
                ipAddr.sin_family = AF_INET;
                ipAddr.sin_addr.s_addr = htonl(INADDR_ANY);
                ipAddr.sin_port = htons((u_short)tcpPort);
                if (bind(server->ipSock,(struct sockaddr *)&ipAddr,
			sizeof(ipAddr)) < 0)
                {
                        perror("Can't start server : IP Bind ");
			exit(1);
                }
                listen(server->ipSock,128);
        }
        else
        {
                server->ipSock = -1;
        }


	/*
	** Setup the UNIX domain socket
	*/
	server->unixPort = strdup(configGetCharEntry("general","unix_port"));
	msqlDebug1(MOD_GENERAL,"UNIX Socket is %s\n", server->unixPort);
	server->unixSock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (server->unixSock < 0)
	{
		perror("Can't start server : UNIX Socket ");
		exit(1);
	}
	bzero(&unixAddr, sizeof(unixAddr));
	unixAddr.sun_family = AF_UNIX;
	strcpy(unixAddr.sun_path, server->unixPort);
	unlink(server->unixPort);
	if( bind(server->unixSock, (struct sockaddr *)&unixAddr,
		sizeof(unixAddr)) < 0)
	{
		perror("Can't start server : UNIX Bind ");
		exit(1);
	}
	listen(server->unixSock,128);
	chmod(server->unixPort,0777);
}



RETSIGTYPE puntServer(sig)
	int	sig;
{
	int	clientSock;

	signal(sig,SIG_IGN);
	clientSock = 3;
	if (sig == -1)
	{
		printf("\n\nNormal Server shutdown!\n\n");
	}
	else
	{
		printf("\nServer Aborting!\n");
	}
	while(clientSock < maxCons + globalServer->config.cacheDescriptors)
	{
		if (globalServer->conArray[clientSock].db)
		{
			printf("Forcing close on Socket %d\n",clientSock);
			shutdown(clientSock,2);
			close(clientSock);
		}
		clientSock++;
	}
	if (globalServer->unixSock >= 0)
	{
		shutdown(globalServer->unixSock,2);
		close(globalServer->unixSock);
		unlink(globalServer->unixPort);
		free(globalServer->unixPort);
	}
	printf("\n");
	cacheDropTableCache();

	if (debugSet(MOD_MALLOC))
	{
		/*
		**
		fprintf(stderr,"\n\nmalloc() leak detection .....\n");
		checkBlocks(MALLOC_BLK);
		**
		*/
	}
	if (debugSet(MOD_MMAP))
	{
		/*
		**
		fprintf(stderr,"\n\nmmap() leak detection .....\n");
		checkBlocks(MMAP_BLK);
		**
		*/
	}
	memDropCaches();
	printf("\n\nmSQL Daemon Shutdown Complete.\n\n");
	if (sig >= 0)
	{
		if(*globalQueryBuf)
		{
			printf("Last query executed : %s\n\n", 
				globalQueryBuf);
		}
		exit(1);
	}
	return;
}




void freeClientConnection(server, sock)
	msqld	*server;
	int	sock;
{
	if(server->conArray[sock].db)
	{
		free(server->conArray[sock].db);
		server->conArray[sock].db = NULL;
	}
	if(server->conArray[sock].host)
	{
		free(server->conArray[sock].host);
		server->conArray[sock].host = NULL;
	}
	if(server->conArray[sock].user)
	{
		free(server->conArray[sock].user);
		server->conArray[sock].user = NULL;
	}
	server->conArray[sock].sock = -1;
	numCons--;
}


RETSIGTYPE puntClient(sig)
	int	sig;
{
	extern	int netCurrentSock;

	if(globalServer->conArray[netCurrentSock].sock < 0)
	{
		/* Must have already been "force closed" */
		return;
	}
	if (sig)
	{
		signal(sig, puntClient);
		msqlDebug1(MOD_GENERAL,
			"Forced close of client on socket %d due to pipe sig\n",
			netCurrentSock);
	}
	else
	{
		msqlDebug1(MOD_GENERAL,
			"Closing client on socket %d\n", netCurrentSock);
	}
	shutdown(netCurrentSock,2);
	close(netCurrentSock);
	FD_CLR(netCurrentSock,&(globalServer->clientFDs));
	freeClientConnection(globalServer, netCurrentSock);
	return;
}


char *calcUptime()
{
	int	days,
		hours,
		mins,
		secs;
	u_int	tmp,
		uptime;
	static 	char msg[80];

	tmp = uptime = time(NULL) - serverStartTime;
	days = tmp / 3600 / 24;
	tmp -= (days * 3600 * 24);
	hours = tmp  / 3600;
	tmp -= hours * 3600;
	mins = tmp / 60;
	tmp -= mins * 60;
	secs = tmp;
	snprintf(msg,80,"%d days, %d hours, %d mins, %d secs",
		days, hours, mins, secs);
	return(msg);
}

void sendServerStats(server, sock)
	msqld	*server;
	int	sock;
{
	struct	passwd *pw;
	int	uid,
		loop,
		cTime;
	u_int	curTime = time(NULL);
	char	buf[10];

	snprintf(packet,PKT_LEN, "%s%s\n%s\n%s\n%s\n\n", 
		"Mini SQL Version ", SERVER_VERSION,
		"Copyright (c) 1993-94 David J. Hughes",
		"Copyright (c) 1995-2012 Hughes Technologies Pty Ltd.",
		"All rights reserved.");
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Config file      : %s\n",confFile);
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Max connections  : %d\n",maxCons);
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Cur connections  : %d\n",numCons);
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Backend processes: %d\n",numKids);
	netWritePacket(sock);

#if defined(_OS_UNIX) || defined(_OS_OS2)
	uid = getuid();
	pw = getpwuid(uid);
	if (pw != NULL)
	{
		snprintf(packet,PKT_LEN,"Running as user  : %s\n", pw->pw_name);
	}
	else
	{
		snprintf(packet,PKT_LEN,"Running as user  : UID %d\n", uid);
	}
	netWritePacket(sock);
#endif

	snprintf(packet,PKT_LEN,"Server uptime    : %s\n", calcUptime());
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Connection count : %d\n", conCount);
	netWritePacket(sock);

	strcpy(packet, "\nConnection table :\n");
	netWritePacket(sock);
	loop = 0;
	strcpy(packet,
	    "  Sock    Username       Hostname        Database    Connect   Idle   Queries\n");
	strcat(packet,
	    " +-----+------------+-----------------+------------+---------+------+--------+\n");
	netWritePacket(sock);
	while(loop < maxCons)
	{
		if (server->conArray[loop].user)
		{
			cTime = (int)(curTime - server->conArray[loop].connectTime)/60;
			snprintf(buf,sizeof(buf),
				"%2dH %2dM",cTime/60, cTime%60);
			snprintf(packet,PKT_LEN,
			 " | %3d | %-10s | %-15s | %-10s | %s | %4d | %6d |\n",
			    loop, server->conArray[loop].user, 
			    server->conArray[loop].host ?
				server->conArray[loop].host:"UNIX Sock",
			    server->conArray[loop].db ? 
				server->conArray[loop].db : "No DB",
			    buf,
			/*
			    (int)(curTime - conArray[loop].lastQuery)/60,
			*/ 
			    0,
			    server->conArray[loop].numQueries);
			netWritePacket(sock);
		}
		loop++;
	}
	strcpy(packet,
	    " +-----+------------+-----------------+------------+---------+------+--------+\n");
	netWritePacket(sock);

	strcpy(packet,"\n");
	netWritePacket(sock);
}


void logQuery(logFP,con, query) 
	FILE	*logFP;
	cinfo_t	*con;
	char	*query;
{
	char	dateBuf[80];
	struct  tm *locTime;
	time_t	clock;

	clock = time(NULL);
        locTime = localtime(&clock);
        strftime(dateBuf,sizeof(dateBuf),"%d-%b-%Y %H:%M:%S",locTime);

	fprintf(logFP,"%s %s %s %s %d\n%s\n",
		dateBuf,
		con->user?con->user:"NO_USER",
		con->host?con->host:"UNIX_SOCK",
		con->db?con->db:"NO_DB",
		query?(int)strlen(query):13,
		query?query:"MISSING_QUERY");
	fflush(logFP);
}


void usage()
{
	printf("Usage :  msql2d [-f ConfFile]\n\n");
}






#define STRN_CPY(d,s,l)   { strncpy((d),(s),(l)); (d)[(l)-1] = 0; }
#define STR_NE(x,y)       (strcasecmp((x),(y)) != 0)
#define STRING_LENGTH     128
   


RETSIGTYPE childHandler(sig)
	int	sig;
{
	int	clientSock;

	signal(SIGCHLD,SIG_IGN);
	printf("\n\nBackend process exited!  Terminating server.\n");
	clientSock = 3;
	while(clientSock < maxCons + globalServer->config.cacheDescriptors)
	{
		if (globalServer->conArray[clientSock].db)
		{
			printf("Forcing close on Socket %d\n",clientSock);
			shutdown(clientSock,2);
			close(clientSock);
		}
		clientSock++;
	}
	if (globalServer->unixSock >= 0)
	{
		shutdown(globalServer->unixSock,2);
		close(globalServer->unixSock);
		unlink(globalServer->unixPort);
		free(globalServer->unixPort);
	}
	/*
	** Close off the children
	*/
	sync();
	fprintf(stderr,"Closing backend connections.\n\n");
	printf("\n");
	exit(1);
}


RETSIGTYPE sigTrap(sig)
	int	sig;
{
	int		clientSock;
	char		path[256];

#ifdef WIN32
	fprintf(stderr,"\nExiting on signal\n\n");
#else
	signal(sig,SIG_IGN);
        fprintf(stderr,"Process %d exiting on a sig %d!\n", getpid(), sig);
#endif
	clientSock = 3;
	msqlDebug0(MOD_ANY,"Forced server shutdown due to bad signal!\n");
	while(clientSock < maxCons + globalServer->config.cacheDescriptors)
	{
		if (globalServer->conArray[clientSock].db)
		{
			printf("Forcing close on Socket %d\n",clientSock);
			shutdown(clientSock,2);
			close(clientSock);
		}
		clientSock++;
	}
	if (globalServer->unixSock >= 0)
	{
		shutdown(globalServer->unixSock,2);
		close(globalServer->unixSock);
		unlink(globalServer->unixPort);
		free(globalServer->unixPort);
	}
	/*
	** Close off the children
	*/
	sync();
	fprintf(stderr,"Closing backend connections.\n\n");
	printf("\n");


#ifdef HAVE_SETRLIMIT
	{
        	struct rlimit   limit;
                limit.rlim_cur = limit.rlim_max = RLIM_INFINITY;
                setrlimit(RLIMIT_CORE, &limit);
	}
#endif
	if (*globalQueryBuf)
	{
		printf("Last query : %s\n\n", globalQueryBuf);
	}
	snprintf(path,255, "%s/core",  INST_DIR);
	mkdir(path, 0666);
	chdir(path);
	abort();
}




void setupSignals()
{
#ifdef SIGSEGV
	signal(SIGSEGV,sigTrap);
#endif
#ifdef SIGBUS
	signal(SIGBUS,sigTrap);
#endif
#ifdef SIGINT
	signal(SIGINT,puntServer);
#endif
#ifdef SIGQUIT
	signal(SIGQUIT,puntServer);
#endif
#ifdef SIGKILL
	signal(SIGKILL,puntServer);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE,puntClient);
#endif
#ifdef SIGTERM
	signal(SIGTERM,puntServer);
#endif
#ifdef SIGHUP
	signal(SIGHUP,SIG_IGN);
#endif
/*
#ifdef SIGCHLD
	signal(SIGCHLD,childHandler);
#endif
*/
}


void processClientRequest(server, client)
	msqld		*server;
	int		client;
{
	mQuery_t	*query,
			dummyQuery;
	int		clientCommand,
			comSock;
	char		dbname[255],
			*arg, *arg2, *arg3,
			*cp,
			tmpChar;



	comSock = server->conArray[client].sock;
	if (netReadPacket(comSock) <= 0)
	{
		msqlDebug1(MOD_GENERAL,"Command read on sock %d failed!\n",
			comSock);
		clientCommand = QUIT;
	}
	else
	{
		clientCommand = atoi(packet);
	}
	msqlDebug3(MOD_GENERAL,"Command on sock %d = %d (%s)\n", comSock, 
		clientCommand, comTable[clientCommand]);

	switch(clientCommand)
	{
		case QUIT:
			msqlDebug0(MOD_GENERAL,"DB QUIT!\n");
			FD_CLR(comSock,&server->clientFDs);
			shutdown(comSock,2);
			close(comSock);
			if(server->conArray[client].user)
			{
				free(server->conArray[client].user);
				server->conArray[client].user = NULL;
			}
			if(server->conArray[client].host)
			{
				free(server->conArray[client].host);
				server->conArray[client].host = NULL;
			}
			if(server->conArray[client].db)
			{
				free(server->conArray[client].db);
				server->conArray[client].db = NULL;
			}
			server->conArray[client].sock = -1;
			numCons--;
			break;

		case INIT_DB:
			cp=(char *)strtok(packet+2,"\n\r");
			if (!cp)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			strcpy(dbname,cp);
			msqlDebug1(MOD_GENERAL,"DBName = %s\n", dbname);
			server->conArray[client].access = aclCheckAccess(
                                        dbname, server->conArray + client);
			if(server->conArray[client].access == NO_ACCESS)
			{
				netError(comSock,ACCESS_DENIED_ERROR);
				break;
			}
			if(server->conArray[client].db)
			{
				free(server->conArray[client].db);
				server->conArray[client].db = NULL;
			}
			server->conArray[client].db = (char *)strdup(dbname);
			if (utilCheckDB(server, dbname) < 0)
			{
				netError(comSock,errMsg);
				break;
			}
			netOK(comSock);
			break;

		case QUERY:
			if (!server->conArray[client].db)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			curSock = comSock;
			cp=(char *)(packet+2);
			arg = cp;
			tmpChar = 0;
			if (strlen(arg) > 4096)
			{
				tmpChar = *(arg+4096);
				*(arg+4096) = 0;
			}
			msqlDebug1(MOD_QUERY,"Query = %s",arg);

			if (tmpChar)
				*(arg+4096) = tmpChar;
			/*aclSetPerms(RW_ACCESS);*/
			aclSetPerms(server->conArray[client].access);
			strncpy(globalQueryBuf,arg,MSQL_GLOBAL_BUF_LEN);
			query = parseQuery(server,arg, comSock, 
				server->conArray[client].user, 
				server->conArray[client].db);
			if (query)
			{
				query->queryTime = time(NULL);
				processQuery(server, query, client, arg);
				parseCleanQuery(query);
			}
			break;

		case EXPLAIN:
			if (!server->conArray[client].db)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			curSock = comSock;
			cp=(char *)(packet+2);
			while(*cp != ':') cp++;
			arg = (char *)strdup(cp + 1);
			tmpChar = 0;
			if (strlen(arg) > 4096)
			{
				tmpChar = *(arg+4096);
				*(arg+4096) = 0;
			}
			msqlDebug1(MOD_QUERY,"Explain = %s",arg);
			if (tmpChar)
				*(arg+4096) = tmpChar;
			query = parseQuery(server, arg, comSock, 
				server->conArray[client].user, 
				server->conArray[client].db);
			if (query)
			{
				query->explainOnly = 1;
				processQuery(server, query, client, arg);
				parseCleanQuery(query);
			}
			if(arg)
				free(arg);
			break;

		case DB_LIST:
			curSock = comSock;
			processListDBs(server, comSock);
			break;

		case SEQ_INFO:
			if (!server->conArray[client].db)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			cp = (char *)index(packet,':');
			cp=(char *)strtok(cp+1, "\n\r");
			arg = (char *)strdup(cp);
			curSock = comSock;
			processSequenceInfo(server, comSock, arg, 
				server->conArray[client].db);
			if(arg)
				free(arg);
			break;

		case TABLE_LIST:
			if (!server->conArray[client].db)
			{
				netError(comSock, NO_DB_ERROR);
				break;
			}
			curSock = comSock;
			processListTables(server, comSock, 
				server->conArray[client].db);
			break;

		case FIELD_LIST:
			if (!server->conArray[client].db)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, "\n\r");
			arg = (char *)strdup(cp);
			curSock = comSock;
			processListFields(server, comSock, arg, 
				server->conArray[client].db);
			if(arg)
				free(arg);
			break;

		case INDEX_LIST:
			if (!server->conArray[client].db)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, ":\n\r");
			arg = (char *)strdup(cp);
			cp=(char *)strtok(NULL,"\n\r");
			arg2 = (char *)strdup(cp);
			curSock = comSock;
			processListIndex(server, comSock, arg2, arg, 
				server->conArray[client].db);
			if(arg)
				free(arg);
			if(arg2)
				free(arg2);
			break;

               	case EXPORT:
			if (!server->conArray[client].db)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			arg = arg2 = arg3 = NULL;
			cp=(char *)strtok(packet+2, ":\n\r");
			arg = (char *)strdup(cp);
			cp=(char *)strtok(NULL,":\n\r");
			arg2 = (char *)strdup(cp);
			cp=(char *)strtok(NULL,":\n\r");
			if (cp)
			{
				arg3 = (char *)strdup(cp);
			}
			curSock = comSock;
			dummyQuery.curUser = server->conArray[client].user;
			dummyQuery.curDB = server->conArray[client].db;
			processExportTable(server, comSock, 
				server->conArray[client].db, arg, arg2, arg3,
				NULL);
			if(arg)
				free(arg);
			if(arg2)
				free(arg2);
			if(arg3)
				free(arg3);
			break;

               	case IMPORT:
			if (!server->conArray[client].db)
			{
				netError(comSock,NO_DB_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, ":\n\r");
			arg = (char *)strdup(cp);
			cp=(char *)strtok(NULL,":\n\r");
			arg2 = (char *)strdup(cp);
			curSock = comSock;
			dummyQuery.curUser = server->conArray[client].user;
			dummyQuery.curDB = server->conArray[client].db;
			processImportTable(server, comSock, 
				server->conArray[client].db, arg, arg2, NULL);
			if(arg)
				free(arg);
			if(arg2)
				free(arg2);
			break;

		case CREATE_DB:
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, "\n\r");
			arg = (char *)strdup(cp);
			processCreateDB(server, comSock, arg);
			if(arg)
				free(arg);
			break;

	 	case COPY_DB:
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, ":\n\r");
			arg = (char *)strdup(cp);
			cp=(char *)strtok(NULL, "\n\r");
			arg2 = (char *)strdup(cp);
			processCopyDB(server, comSock,arg, arg2);
			if(arg)
				free(arg);
			if(arg2)
				free(arg2);
			break;

		case MOVE_DB:
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, ":\n\r");
			arg = (char *)strdup(cp);
			cp=(char *)strtok(NULL, "\n\r");
			arg2 = (char *)strdup(cp);
			processMoveDB(server, comSock,arg, arg2);
			if(arg)
				free(arg);
			if(arg2)
				free(arg2);
			break;

		case DROP_DB:	
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, "\n\r");
			arg = (char *)strdup(cp);
			processDropDB(server,comSock,arg);
			if(arg)
				free(arg);
			break;

		case RELOAD_ACL:
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			aclReloadFile(comSock);
			netEndOfList(comSock);
			break;


		case SHUTDOWN:
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			netEndOfList(comSock);
			puntServer(-1);
			exit(0);
			break;

		case SERVER_STATS:
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			sendServerStats(server,comSock);
			netEndOfList(comSock);
			break;

		case CHECK_TABLE:	
			if (!aclCheckLocal(&server->conArray[client]))
			{
				netError(comSock,PERM_DENIED_ERROR);
				break;
			}
			cp=(char *)strtok(packet+2, ":\n\r");
			arg = (char *)strdup(cp);
			cp=(char *)strtok(NULL, "\n\r");
			arg2 = (char *)strdup(cp);
			processCheckTable(server,comSock,arg,arg2);
			if(arg)
				free(arg);
			if(arg2)
				free(arg2);
			break;

		default:
			netError(comSock, UNKNOWN_COM_ERROR);
			break;
	}
}


void debugTrap()
{
	/* We use this as a debugger breakpoint */
}

void childStartup()
{
	/* 
	** The broker code references this so we need an empty
	** function for the linker even though it'll never be called
	*/
}


int main(argc,argv)
	int	argc;
	char	*argv[];
{
	msqld	server;
	int	sock,
		newSock,
		maxSock,
		cacheSize,
		opt,
		error,
		errFlag = 0,
		count,
		c;
	char	*uname,
		path[MSQL_PATH_LEN],
		pidBuf[10];
	int	pidFD;
	fd_set	readFDs;

        extern  char *optarg;
#ifdef HAVE_SETRLIMIT
	struct	rlimit limit;
#endif
	struct	passwd *pwd;
	struct	stat sbuf;

#ifdef	DEBUG_BSD_MALLOC
	extern char *malloc_options;
	malloc_options = "AJ";
#endif
#ifdef 	DEBUG_MEMTRACE
	MemTrace_Init (argv[0], MEMTRACE_REPORT_ON_EXIT);
#endif

	/*
	** Turn off IO buffering so that stdlib doesn't
	** eat 64k blocks of memory
	*/
	setbuf(stdout, NULL);
	setbuf(stdin, NULL);

	bzero(&server, sizeof(server));
	globalServer = &server;

	serverStartTime = time(NULL);
        printf("\n\nMini SQL Version %s\n",SERVER_VERSION);
        printf("Copyright (c) 1993-94 David J. Hughes\n");
        printf("Copyright (c) 1995-2012 Hughes Technologies Pty Ltd.\n");
        printf("All rights reserved.\n\n");

        /* Try to set the locale but don't bail if it fails (freeBSD) */
        setlocale(LC_ALL, "");

	/*
	** Handle the command line args
	*/
	*confFile = 0;
        while((c=getopt(argc,argv,"f:"))!= -1)
        {
                switch(c)
                {
                        case 'f':
                                if (*confFile)
                                        errFlag++;
                                else
                                        strcpy(confFile,optarg);
                                break;
                        case '?':
                                errFlag++;
                                break;
                }
	}
	if (errFlag)
	{
		usage();
		exit(1);
	}

	if (*confFile == 0)
	{
		snprintf(confFile,sizeof(confFile),"%s/msql.conf", INST_DIR);
	}
	printf("\tLoading configuration from '%s'.\n",confFile);
	configLoadFile(confFile);
	server.config.instDir=strdup(configGetCharEntry("general","inst_dir"));
	server.config.dbDir = strdup(configGetCharEntry("general", "db_dir"));
	server.config.readOnly = configGetIntEntry("system", "read_only");
	server.config.msyncTimer = configGetIntEntry("system", "msync_timer");
	server.config.sortMaxMem = configGetIntEntry("system","sort_max_mem");
	server.config.tableCache = configGetIntEntry("system","table_cache");
	server.config.cacheDescriptors = server.config.tableCache * 
		(NUM_INDEX + 2);
	server.config.needFileLock = 0;
	server.config.hasBroker = 0;

	/*
	** Check out the file descriptor limit
	**
	** Some boxes have broken getrlimit() implementations (or
	** missing RLIMIT_NOFILE) so we try to use sysconf first
	**
	** Watch out for broken BSDI here.  It looks like BSDI returns
	** the kernel MAX_FILES not the per process MAX_FILE in
	** getrlimit() which kinda sucks if you want to use select()!
	*/

#	ifdef HAVE_RLIMIT_NOFILE
		getrlimit(RLIMIT_NOFILE,&limit);
		limit.rlim_cur = (limit.rlim_max > 256)? 256 : limit.rlim_max;
		setrlimit(RLIMIT_NOFILE,&limit);
		getrlimit(RLIMIT_NOFILE,&limit);
		maxCons = limit.rlim_cur - server.config.cacheDescriptors;
		printf("\tServer process reconfigured to accept %d ",maxCons);
		printf("connections.\n");
#	else
#	ifdef HAVE_SYSCONF
		maxCons = sysconf(_SC_OPEN_MAX)-server.config.cacheDescriptors;
		printf("\tServer process reconfigured to accept %d ",maxCons);
		printf("connections.\n");
#  	else
		maxCons = 64 - server.config.cacheDescriptors;
		printf("\tDynamic server config failed!\n");
		printf("\tServer can accept %d connections.\n", maxCons);
#	endif
#	endif
	cacheSize = configGetIntEntry("system", "table_cache");
	if (cacheSize >= 2)
		printf("\tServer table cache holds %d entries.\n", cacheSize);
	else
		printf("\tServer table cache config too small.  Using 2.\n");

	if (maxCons > 254)
		maxCons = 254;

	/*
	** Set up query logging if required
	*/
        server.logFP = NULL;
        if (configGetIntEntry("system","query_log"))
        {
                server.logFP = fopen(configGetCharEntry("system",
			"query_log_file"),"a");
                if (server.logFP)
		{
                        printf("\tQuery logging enabled.\n");
			setbuf(server.logFP, NULL);
		}
                else
		{
                        printf("\tERROR : Open of query log failed!\n");
		}
	}

        server.updateFP = NULL;
        if (configGetIntEntry("system","update_log"))
        {
                server.updateFP = fopen(configGetCharEntry("system",
			"update_log_file"),"a");
                if (server.updateFP)
		{
                        printf("\tUpdate logging enabled.\n");
			setbuf(server.updateFP, NULL);
		}
                else
		{
                        printf("\tERROR : Open of update log failed!\n");
		}
	}

	/*
	** We have to do this now.  If msql is configured to run as nobody
	** then nobody would have to have write perms for the INST_DIR
	** This way we can start as root and then run as the MSQL_USER when
	** we're setup properly
	*/
	initDebug(1);
	netInitialise();
	initServer(&server);

	(void)snprintf(path, MSQL_PATH_LEN, "%s", 
		(char *)configGetCharEntry("general","pid_file"));
	pidFD = open(path, O_CREAT|O_TRUNC|O_RDWR, 0700);
	if (pidFD < 0)
	{
		perror("Error : Couldn't open PID file");
		exit(1);
	}
	else
	{
		if (lockNonBlockingLock(pidFD) < 0)
		{
			printf("\nError : Cannot lock PID file!\n\n");
			exit(1);
		}
		sprintf(pidBuf,"%d", (int)getpid());
		write(pidFD, pidBuf, strlen(pidBuf) + 1);
	}


	/*
	** Are we running as the right user?
	*/
	pwd = getpwuid(getuid());
	if (!pwd)
	{
		printf("\nError!  No username for our UID (%d)\n\n",
			(int)getuid());
		exit(1);
	}
	uname = (char *)configGetCharEntry("general","msql_user");
	if (strcmp(uname,pwd->pw_name) != 0)
	{
		pwd = getpwnam(uname);
		if (!pwd)
		{
			printf("\nError!  Unknown username (%s)\n\n", uname);
			exit(1);
		}
		chown(path,pwd->pw_uid, pwd->pw_gid);
		chown("/proc/self/fd",pwd->pw_uid,pwd->pw_gid);
		if (setuid(pwd->pw_uid) < 0)
		{
			printf("\nError!  Can't run as user '%s'\n\n",uname);
			exit(1);
		}
	}
	printf("\tServer running as user '%s'.\n",uname);
	printf("\tServer mode is %s.\n", configGetIntEntry("system",
		"read_only")?"Read-Only":"Read/Write");

	/*
	** Ensure that the correct user owns the database files
	*/
	snprintf(path,MSQL_PATH_LEN, "%s/msqldb",
		(char *)configGetCharEntry("general","inst_dir"));
	if (stat(path,&sbuf) < 0)
	{
		printf("\nError!  Can't stat '%s'\n\n",path);
		exit(1);
	}

	if (sbuf.st_uid != getuid())
	{
		printf("\nError!  '%s' is not owned by '%s'\n\n",path,
			pwd->pw_name);
		exit(1);
	}

	/* Ensure there's no wierd FD's open */
	for ( count = server.unixSock + 1; count<= 255; count++)
	{
		close(count);
	}

        /* Try to set the locale but don't bail if it fails (freeBSD) */
        setlocale(LC_ALL, "");

        /*
        ** Initialise the connection array
        */
        count = 0;
        while (count < 255)
        {
                server.conArray[count].sock = -1;
                server.conArray[count].db = NULL;
                count++;
        }


	/*
	** OK, on with the show
	*/

	umask(0);
	numCons=0;
	maxSock = 0;
	/* chdir(msqlHomeDir); */
	tableCleanTmpDir(&server);

#	ifdef BETA_RELEASE
	timeRemain = (863060000 + 60*24*60*60) - time(NULL);
	if (timeRemain < 0)
	{
		printf("\nThis Beta test release has expired!\n");
		printf("Check out www.Hughes.com.au for a new release!\n\n");
		exit(1);
	}
	else
	{
		printf("\n\tThis Beta test release expires in %d days.\n\n",
			(int)(timeRemain / (24*60*60)));
	}
#	endif

	chmod(path,0700);
	umask(0077);
	setupSignals();
	(void)bzero(&server.conArray,sizeof(&server.conArray));
	aclLoadFile(1);
	msqlDebug0(MOD_GENERAL,
		"miniSQL debug mode.  Waiting for connections.\n");

        FD_ZERO(&server.clientFDs);
	while(1)
	{
               	bcopy(&server.clientFDs, &readFDs, sizeof(readFDs));
		if (server.unixSock >= 0)
               		FD_SET(server.unixSock,&readFDs);
		if (server.ipSock >= 0)
               		FD_SET(server.ipSock,&readFDs);
                if(select(255,&readFDs,0,0,0) < 0)
                {
                       	if (errno == EINTR)
			{
				eintrCount++;
				if(eintrCount > 100)
				{
					printf(SELECT_EINTR_ERRORS);
					puntServer(0);
				}
                               	continue;
			}
                       	puntServer(0);
                }
		eintrCount = 0;

		/*
		** How about a new connection?  Grab the lock so we
		** can ensure the broker notification gets out before
		** the new connection becomes active
		*/
		sock = 0;
		if (server.unixSock >= 0 && FD_ISSET(server.unixSock,&readFDs))
			sock = server.unixSock;
		if (server.ipSock >= 0 && FD_ISSET(server.ipSock,&readFDs))
			sock = server.ipSock;

		if (sock != 0)
		{
                        struct  sockaddr_un     cAddr;
                        int  cAddrLen;

			if (FD_ISSET(server.unixSock,&readFDs))
				sock = server.unixSock;
			else
				sock = server.ipSock;

			bzero(&cAddr, sizeof(cAddr));
			cAddrLen = sizeof(struct sockaddr_un);
			newSock = accept(sock, (struct sockaddr *)&cAddr, 
				&cAddrLen);
			if(newSock < 0)
			{
				perror("Error in accept ");
				puntServer(-1);
				exit(1);
			}
			if (newSock > maxSock)
				maxSock = newSock;
			if (server.conArray[newSock].db)
			{
				if(server.conArray[newSock].db)
				{
					free(server.conArray[newSock].db);
					server.conArray[newSock].db = NULL;
				}
				if(server.conArray[newSock].host)
				{
					free(server.conArray[newSock].host);
					server.conArray[newSock].host = NULL;
				}
				if(server.conArray[newSock].user)
				{
					free(server.conArray[newSock].user);
					server.conArray[newSock].user = NULL;
				}
			}

			/*
			** Are we over the connection limit
			*/
			numCons++;
			if (numCons > maxCons)
			{
				numCons--;
				netError(newSock,CON_COUNT_ERROR);
				shutdown(newSock,2);
				close(newSock);
				continue;
			}


			/*
			** store the connection details
			*/

			msqlDebug1(MOD_GENERAL,
				"New connection received on %d\n", newSock);
			error = 0;

                        if (sock == server.ipSock)
                        {
                                struct  hostent *hp;
				struct	sockaddr_in remote;
				char	*ipAddr;
                                int  	addrLen;

                                addrLen = sizeof(struct sockaddr);
                                getpeername(newSock, (struct sockaddr *)
                                        &remote, &addrLen);
                                ipAddr = inet_ntoa(remote.sin_addr);
                                if (ipAddr)
                                {
                                        server.conArray[newSock].clientIP =
                                                strdup(ipAddr);
                                }
                                if (configGetIntEntry("system","host_lookup")
					== 1)
                                {
                                        /*
                                        ** Validate remote host
                                        */

                                        if(auth_hostname(&remote))
                                        {
                                                netError(newSock,
                                                        BAD_HOST_ERROR);
                                                error = 1;
                                                numCons--;
                                                shutdown(newSock,2);
                                                close(newSock);
                                        }

                                        /*
                                        ** Grab the hostname for later
                                        */
                                        hp = (struct hostent *)gethostbyaddr(
                                                (char *)&remote.sin_addr,
                                                sizeof(remote.sin_addr),
                                                AF_INET);
                                }
                                else
                                {
                                        hp = NULL;
                                }

                                if (!hp)
                                {
                                        server.conArray[newSock].host=(char *)
                                                strdup("UNKNOWN_HOST");
                                        msqlDebug0(MOD_GENERAL,
                                                "Host=UNKNOWN_HOST\n");
                                }
                                else
                                {
                                        server.conArray[newSock].host=(char *)
                                                strdup((char *)hp->h_name);
                                        msqlDebug1(MOD_GENERAL,"Host = %s\n",
                                                hp->h_name);
                                }
                        }
                        else
                        {
                                server.conArray[newSock].host = NULL;
                                server.conArray[newSock].clientIP = NULL;
                                msqlDebug0(MOD_GENERAL,"Host = UNIX domain\n");
                        }

			server.conArray[newSock].connectTime = 0;
			server.conArray[newSock].numQueries = 0;

			if (!error)
			{
				opt=1;
				setsockopt(newSock,SOL_SOCKET,SO_KEEPALIVE,
					(char *) &opt, sizeof(opt));
				/* fcntl(newSock, F_SETFL, O_NONBLOCK);*/
				snprintf(packet, PKT_LEN,
					"0:%d:%s\n",
					PROTOCOL_VERSION,SERVER_VERSION);
				netWritePacket(newSock);
				if (netReadPacket(newSock) <=0)
				{
					netError(newSock,HANDSHAKE_ERROR);
					numCons--;
					shutdown(newSock,2);
					close(newSock);
					server.conArray[newSock].host=NULL;
					server.conArray[newSock].clientIP=NULL;
				}
				else
				{
					FD_SET(newSock, &server.clientFDs);
					uname = (char *)strtok(packet,"\n");
					msqlDebug1(MOD_GENERAL,
						"User = %s\n",uname);
					if(server.conArray[newSock].user)
					    free(server.conArray[newSock].user);
					server.conArray[newSock].user = (char *)
						strdup(uname);
					server.conArray[newSock].sock =
						newSock;
					netEndOfList(newSock);
				}
			}
			continue;
		}

		/*
		** Looks like a request from an existing client
		*/
		count = 0;
		while(count <= maxSock)
		{
			if (server.conArray[count].sock == -1)
			{
				count++;
				continue;
			}
			if (!FD_ISSET(server.conArray[count].sock,&readFDs))
			{
				count++;
				continue;
			}
			processClientRequest(&server, count);
			count++;
		}
	}
}
