/*
** Copyright (c) 1995-2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  
**
** Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: main_broker.c,v 1.25 2012/06/04 05:40:02 bambi Exp $
**
*/

/*
** Module	: main : msqld_main
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/



/**************************************************************************
** STANDARD INCLUDES
**************************************************************************/

#include <common/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif


/**************************************************************************
** MODULE SPECIFIC INCLUDES
**************************************************************************/

#include <fcntl.h>
#include <limits.h>
#include <locale.h>
#include <errno.h>
#include <pwd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/un.h>
#include <signal.h>

#ifdef HAVE_SETRLIMIT
#  include <sys/resource.h>
#endif

#if HAVE_SYS_MMAN_H
#  include <sys/mman.h>
#endif

#include <common/msql_defs.h>
#include <common/debug/debug.h>
#include <common/portability.h>
#include <common/config/config.h>
#include <msqld/index/index.h>
#include <msqld/includes/msqld.h>
#include <msqld/includes/errmsg.h>
#include <msqld/main/main.h>
#include <msqld/broker/broker.h>
#include <msqld/main/version.h>
#include <msqld/main/process.h>
#include <msqld/main/table.h>
#include <msqld/main/net.h>
#include <msqld/main/acl.h>
#include <msqld/main/cache.h>
#include <msqld/main/util.h>
#include <msqld/main/parse.h>
#include <msqld/main/tcp.h>
#include <msqld/lock/lock.h>


/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/

msqld	*globalServer;

/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/

static void _zeroMessageStruct(msg)
	mMsg_t	*msg;
{
	msg->command = msg->access = msg->client = 0;
	*msg->db = *msg->table = *msg->user = *msg->client_ip = 0;
}
	

/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/


extern  u_char  *yytext,
		*yyprev;
extern	int	yylineno;
extern	int	yydebug;

extern	char	errMsg[];

int	curSock,
	IPsock = 0,
	UNIXsock = 0,
	numCons = 0,
	maxCons = 0,
	conCount = 0,
	numKids = 0,
	eintrCount = 0,
	msyncTimer,
	startTime,
	explainOnly;

FILE	*logFP;

u_int	serverStartTime = 0,
	serverNumCons = 0,
	serverNumQueries = 0;

char	*unixPort,
	confFile[MSQL_PATH_LEN];

cinfo_t	conArray[256];

extern	char	*packet;

char	PROGNAME[] = "msqld",
	BLANK_ARGV[] = "                                                  ";


static char *comTable[] = {
        "???", "Quit", "Init DB", "Query", "DB List", "Table List",
        "Field List", "Create DB", "Drop DB", "Reload ACL",
        "Shutdown", "Index List", "Stats", "Seq Info", "Move DB",
	"Copy DB", "???" };


#if defined(_OS_WIN32)
RETSIGTYPE sigTrap(int sig);
RETSIGTYPE puntServer(int sig);
RETSIGTYPE puntClient(int sig);
#endif



/****************************************************************************
** 	_initServer
**
**	Purpose	: 
**	Args	: 
**	Returns	: 
**	Notes	: 
*/

void initServer()
{
	int	tcpPort,
		opt;
	struct	sockaddr_in	IPaddr;

#ifdef HAVE_SYS_UN_H
	struct	sockaddr_un	UNIXaddr;
#endif

	cacheSetupTableCache();

	/*
	** Create an IP socket
	*/
	if (configGetIntEntry("system", "remote_access"))
	{
		tcpPort = configGetIntEntry("general", "tcp_port");
		msqlDebug1(MOD_GENERAL,"IP Socket is %d\n",tcpPort);
		IPsock = socket(AF_INET, SOCK_STREAM, 0);
		if (IPsock < 0)
		{
			perror("Can't start server : IP Socket ");
			exit(1);
		}
#		ifdef SO_REUSEADDR
		opt = 1;
		setsockopt(IPsock, SOL_SOCKET, SO_REUSEADDR, (char *)&opt,
			sizeof(int));
#		endif

		bzero(&IPaddr, sizeof(IPaddr));
		IPaddr.sin_family = AF_INET;
		IPaddr.sin_addr.s_addr = htonl(INADDR_ANY);
		IPaddr.sin_port = htons((u_short)tcpPort);
		if (bind(IPsock,(struct sockaddr *)&IPaddr,sizeof(IPaddr)) < 0)
		{
			perror("Can't start server : IP Bind ");
			exit(1);
		}
		listen(IPsock,128);
	}
	else
	{
		IPsock = -1;
	}

#	ifdef HAVE_SYS_UN_H
	/*
	** Create the UNIX socket
	*/
	if (configGetIntEntry("system", "local_access"))
	{
		unixPort = (char *)strdup((char *)configGetCharEntry("general", 
			"unix_port"));
	
		msqlDebug1(MOD_GENERAL,"UNIX Socket is %s\n",unixPort);
		UNIXsock = socket(AF_UNIX, SOCK_STREAM, 0);
		if (UNIXsock < 0)
		{
			perror("Can't start server : UNIX Socket ");
			exit(1);
		}
		bzero(&UNIXaddr, sizeof(UNIXaddr));
		UNIXaddr.sun_family = AF_UNIX;
		strcpy(UNIXaddr.sun_path, unixPort);
		unlink(unixPort);
		if (bind(UNIXsock, (struct sockaddr *)&UNIXaddr, 
			sizeof(UNIXaddr)) < 0)
		{
			perror("Can't start server : UNIX Bind ");
			exit(1);
		}
		listen(UNIXsock,128);
		chmod(unixPort,0777);
	}
	else
#	endif
	{
		UNIXsock = -1;
	}
}



RETSIGTYPE puntServer(sig)
	int	sig;
{
	int	clientSock;

	signal(sig,SIG_IGN);
	clientSock = 3;
	if (sig == -1)
	{
		printf("\n\nNormal Server shutdown!\n\n");
	}
	else
	{
		printf("\nServer Aborting!\n");
	}
	while(clientSock < maxCons + globalServer->config.cacheDescriptors)
	{
		if (conArray[clientSock].db)
		{
			printf("Forcing close on Socket %d\n",clientSock);
			shutdown(clientSock,2);
			close(clientSock);
		}
		clientSock++;
	}
	if (IPsock >= 0)
	{
		shutdown(IPsock,2);
		close(IPsock);
	}

	if (UNIXsock >= 0)
	{
		shutdown(UNIXsock,2);
		close(UNIXsock);
		unlink(unixPort);
		free(unixPort);
	}
	printf("\n");
	cacheDropTableCache();

	if (debugSet(MOD_MALLOC))
	{
		/*
		**
		fprintf(stderr,"\n\nmalloc() leak detection .....\n");
		checkBlocks(MALLOC_BLK);
		**
		*/
	}
	if (debugSet(MOD_MMAP))
	{
		/*
		**
		fprintf(stderr,"\n\nmmap() leak detection .....\n");
		checkBlocks(MMAP_BLK);
		**
		*/
	}
	printf("\n\nmSQL Daemon Shutdown Complete.\n\n");
	if (sig >= 0)
	{
		exit(1);
	}
	return;
}




void freeClientConnection(server, sock)
	msqld	*server;
	int	sock;
{
	if(conArray[sock].db)
	{
		free(conArray[sock].db);
		conArray[sock].db = NULL;
	}
	if(conArray[sock].host)
	{
		free(conArray[sock].host);
		conArray[sock].host = NULL;
	}
	if(conArray[sock].user)
	{
		free(conArray[sock].user);
		conArray[sock].user = NULL;
	}
	conArray[sock].sock = -1;
	numCons--;
}


RETSIGTYPE puntClient(sig)
	int	sig;
{
	mMsg_t	message;
	extern	int netCurrentSock;

	if(conArray[netCurrentSock].sock < 0)
	{
		/* Must have already been "force closed" */
		return;
	}
	if (sig)
	{
		signal(sig, puntClient);
		msqlDebug1(MOD_GENERAL,
			"Forced close of client on socket %d due to pipe sig\n",
			netCurrentSock);
		_zeroMessageStruct(&message);
		message.command = CMD_CLIENT_CLOSE;
		message.client = netCurrentSock;
		brokerChildSendMessage(&message);
	}
	else
	{
		msqlDebug1(MOD_GENERAL,
			"Closing client on socket %d\n", netCurrentSock);
	}
	freeClientConnection(NULL, netCurrentSock);
	return;
}


char *calcUptime()
{
	int	days,
		hours,
		mins,
		secs;
	u_int	tmp,
		uptime;
	static 	char msg[80];

	tmp = uptime = time(NULL) - serverStartTime;
	days = tmp / 3600 / 24;
	tmp -= (days * 3600 * 24);
	hours = tmp  / 3600;
	tmp -= hours * 3600;
	mins = tmp / 60;
	tmp -= mins * 60;
	secs = tmp;
	snprintf(msg,80,"%d days, %d hours, %d mins, %d secs",
		days, hours, mins, secs);
	return(msg);
}

void sendServerStats(sock)
	int	sock;
{
	struct	passwd *pw;
	int	uid,
		loop,
		cTime;
	u_int	curTime = time(NULL);
	char	buf[10];

	snprintf(packet,PKT_LEN, "%s%s\n%s\n%s\n%s\n\n", 
		"Mini SQL Version ", SERVER_VERSION,
		"Copyright (c) 1993-94 David J. Hughes",
		"Copyright (c) 1995-2012 Hughes Technologies Pty Ltd.",
		"All rights reserved.");
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Config file      : %s\n",confFile);
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Max connections  : %d\n",maxCons);
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Cur connections  : %d\n",numCons);
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Backend processes: %d\n",numKids);
	netWritePacket(sock);

#if defined(_OS_UNIX) || defined(_OS_OS2)
	uid = getuid();
	pw = getpwuid(uid);
	if (pw != NULL)
	{
		snprintf(packet,PKT_LEN,"Running as user  : %s\n", pw->pw_name);
	}
	else
	{
		snprintf(packet,PKT_LEN,"Running as user  : UID %d\n", uid);
	}
	netWritePacket(sock);
#endif

	snprintf(packet,PKT_LEN,"Server uptime    : %s\n", calcUptime());
	netWritePacket(sock);
	snprintf(packet,PKT_LEN,"Connection count : %d\n", conCount);
	netWritePacket(sock);

	strcpy(packet, "\nConnection table :\n");
	netWritePacket(sock);
	loop = 0;
	strcpy(packet,
	    "  Sock    Username       Hostname        Database    Connect   Idle   Queries\n");
	strcat(packet,
	    " +-----+------------+-----------------+------------+---------+------+--------+\n");
	netWritePacket(sock);
	while(loop < maxCons)
	{
		if (conArray[loop].user)
		{
			cTime = (int)(curTime - conArray[loop].connectTime)/60;
			snprintf(buf,sizeof(buf),
				"%2dH %2dM",cTime/60, cTime%60);
			snprintf(packet,PKT_LEN,
			 " | %3d | %-10s | %-15s | %-10s | %s | %4d | %6d |\n",
			    loop, conArray[loop].user, 
			    conArray[loop].host?conArray[loop].host:"UNIX Sock",
			    conArray[loop].db ? conArray[loop].db : "No DB",
			    buf,
			/*
			    (int)(curTime - conArray[loop].lastQuery)/60,
			*/ 
			    0,
			    conArray[loop].numQueries);
			netWritePacket(sock);
		}
		loop++;
	}
	strcpy(packet,
	    " +-----+------------+-----------------+------------+---------+------+--------+\n");
	netWritePacket(sock);

	strcpy(packet,"\n");
	netWritePacket(sock);
}


void logQuery(logFP,con, query) 
	FILE	*logFP;
	cinfo_t	*con;
	char	*query;
{
	char	dateBuf[80];
	struct  tm *locTime;
	time_t	clock;

	clock = time(NULL);
        locTime = localtime(&clock);
        strftime(dateBuf,sizeof(dateBuf),"%d-%b-%Y %H:%M:%S",locTime);

	fprintf(logFP,"%s %s %s %s %d\n%s\n",
		dateBuf,
		con->user?con->user:"NO_USER",
		con->host?con->host:"UNIX_SOCK",
		con->db?con->db:"NO_DB",
		query?(int)strlen(query):13,
		query?query:"MISSING_QUERY");
	fflush(logFP);
}


void usage()
{
	printf("Usage :  msql2d [-f ConfFile]\n\n");
}



RETSIGTYPE childHandler(sig)
	int	sig;
{
	int	clientSock;

	signal(SIGCHLD,SIG_IGN);
	printf("\n\nBackend process exited!  Terminating server.\n");
	clientSock = 3;
	while(clientSock < maxCons + globalServer->config.cacheDescriptors)
	{
		if (conArray[clientSock].db)
		{
			printf("Forcing close on Socket %d\n",clientSock);
			shutdown(clientSock,2);
			close(clientSock);
		}
		clientSock++;
	}
	if (IPsock >= 0)
	{
		shutdown(IPsock,2);
		close(IPsock);
	}
	if (UNIXsock >= 0)
	{
		shutdown(UNIXsock,2);
		close(UNIXsock);
		unlink(unixPort);
		free(unixPort);
	}
	/*
	** Close off the children
	*/
	sync();
	fprintf(stderr,"Closing backend connections.\n\n");
	brokerCloseChildren();
	printf("\n");
	exit(1);
}


RETSIGTYPE sigTrap(sig)
	int	sig;
{
	int	clientSock;
	char	path[256];

#ifdef WIN32
	fprintf(stderr,"\nHit by signal\n\n");
#else
	signal(sig,SIG_IGN);
        fprintf(stderr,"Process %d exiting on a sig %d!\n", getpid(), sig);
#endif
	clientSock = 3;
	msqlDebug0(MOD_ANY,"Forced server shutdown due to bad signal!\n");
	while(clientSock < maxCons + globalServer->config.cacheDescriptors)
	{
		if (conArray[clientSock].db)
		{
			printf("Forcing close on Socket %d\n",clientSock);
			shutdown(clientSock,2);
			close(clientSock);
		}
		clientSock++;
	}
	if (IPsock >= 0)
	{
		shutdown(IPsock,2);
		close(IPsock);
	}
	if (UNIXsock >= 0)
	{
		shutdown(UNIXsock,2);
		close(UNIXsock);
		unlink(unixPort);
		free(unixPort);
	}
	/*
	** Close off the children
	*/
	sync();
	fprintf(stderr,"Closing backend connections.\n\n");
	brokerCloseChildren();
	printf("\n");

#ifdef HAVE_SETRLIMIT  
        {
                struct rlimit   limit;
                limit.rlim_cur = limit.rlim_max = RLIM_INFINITY;
                setrlimit(RLIMIT_CORE, &limit); 
        }
#endif
        snprintf(path,255, "%s/core",  INST_DIR);
        mkdir(path, 0666);
        chdir(path); 
        abort();
}




void setupSignals()
{
#ifdef SIGSEGV
	signal(SIGSEGV,sigTrap);
#endif
#ifdef SIGBUS
	signal(SIGBUS,sigTrap);
#endif
#ifdef SIGINT
	signal(SIGINT,puntServer);
#endif
#ifdef SIGQUIT
	signal(SIGQUIT,puntServer);
#endif
#ifdef SIGKILL
	signal(SIGKILL,puntServer);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE,puntClient);
#endif
#ifdef SIGTERM
	signal(SIGTERM,puntServer);
#endif
#ifdef SIGHUP
	signal(SIGHUP,SIG_IGN);
#endif
#ifdef SIGCHLD
	signal(SIGCHLD,childHandler);
#endif
}


void childStartup(server)
	msqld		*server;
{
	mQuery_t	*query,
			dummyQuery;
	int		count,
			clientCommand,
			remain,
			numBytes,
			offset,
			comSock,
			endOfQueue,
			haveLock,
			clientsProcessed;
	char		dbname[255],
			*arg, *arg2, *arg3,
			*cp,
			go,
			tmpChar;
        fd_set  	readFDs,
			clientFDs;
	mMsg_t		message;
	struct timeval	timeout;



	/* Ensure there's no wierd FD's open */
	for ( count = BROKER_OOB_FD + 1; count<=255; count++)
	{
		close(count);
	}

        /* Try to set the locale but don't bail if it fails (freeBSD) */
        setlocale(LC_ALL, "");


        /*
        ** Initialise the connection array
        */
        count = 0;
        while (count < 255)
        {
                conArray[count].sock = -1;
                conArray[count].db = NULL;
                count++;
        }


	/*
	** OK, on with the show
	*/

	bzero(&dummyQuery, sizeof(dummyQuery));
	strcpy(PROGNAME,"child");
	yytext = NULL;

        umask(0077);
        setupSignals();
        aclLoadFile(0);
        (void)bzero(&readFDs,sizeof(fd_set));
        msqlDebug0(MOD_GENERAL,
                "miniSQL debug mode.  Waiting for connections.\n");

        FD_ZERO(&clientFDs);
	while(1)
	{
                /*
		** Before we waste cycles getting the IPC lock (which
		** we only need if we select() on the client connection
		** sockets), check to see if there's anything waiting
		** on the broker socket (which is dedicated to us and
		** we can do with as we please).
		*/
		haveLock = 0;
		FD_ZERO(&readFDs);
                FD_SET(BROKER_FROM_FD,&readFDs);
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
                if(select(255,&readFDs,0,0,&timeout) == 1)
		{
			/*
			** The broker has jabbed us.  Run the queue
			** without worrying about locking for access
			** to the client sockets.  We won't touch them
			** this time.
			*/
			msqlDebug0(MOD_GENERAL,
				"Found something on the broker sock\n");
		}
		else
		{
			/*
			** OK, a poll of the broker socket didn't get
			** us anywhere.  Grab the lock and check out all
			** the descriptors
			*/
                	bcopy(&clientFDs, &readFDs, sizeof(readFDs));
                	FD_SET(BROKER_FROM_FD,&readFDs);
                	lockGetIpcLock(server);
			haveLock = 1;
                	if(select(255,&readFDs,0,0,0) < 0)
                	{
                        	if (errno == EINTR)
                                	continue;
                        	puntServer(0);
                	}
			msqlDebug0(MOD_GENERAL,"Select broken\n");
		}

                /*
                ** Is there something in the broker's queue?
                */
                if (FD_ISSET(BROKER_FROM_FD,&readFDs))
                {
			/*
			** Hmmm, something came in while we were
			** sleeping in the select.
			**
			** The broker has jabbed us.  Drop the lock,
			** read the jab byte, and ask it for any 
			** queued messages.
			*/

			if (haveLock)
			{
                        	lockReleaseIpcLock(server);
				haveLock = 0;
			}

			read(BROKER_FROM_FD,&go,1);

			_zeroMessageStruct(&message);
			message.command = CMD_RUN_QUEUE;
			message.client = 0;
			brokerChildSendMessage(&message);

			endOfQueue = 0;
                        while(endOfQueue == 0)
                        {
                                remain = sizeof(mMsg_t);
                                offset = 0;
                                while(remain)
                                {
                                        cp = ((char *)&message) + offset;
                                        numBytes = read(BROKER_FROM_FD,cp,
						remain);
                                        if (numBytes <= 0)
                                                puntServer(SIGQUIT);
                                        remain -= numBytes;
                                        offset += numBytes;
                                }


				switch(message.command)
				{
				case CMD_QUEUE_END:
                                        msqlDebug0(MOD_BROKER,
                                            "Processing Broker QUEUE_END\n");
					endOfQueue = 1;
					break;

				case CMD_CLIENT_OPEN:
					comSock = brokerRecvFD(BROKER_FROM_FD);
                                        if (comSock < 1)
                                        {
                                                puntServer(0);
                                        }
                                        msqlDebug0(MOD_BROKER,
                                            "Processing Broker CLIENT_OPEN\n");
                                        msqlDebug2(MOD_BROKER,
                                            "Got new connection (%d on %d)\n",
                                            message.client, comSock);
                                        FD_SET(comSock, &clientFDs);
                                        conArray[message.client].sock = comSock;
                                        conArray[message.client].user =
                                                (char *)strdup(message.user);
					conArray[message.client].connectTime =
						time(NULL);
					conCount++;
					numCons++;
					break;

                                
                                case CMD_FLUSH_CACHE:
                                        msqlDebug2(MOD_BROKER,
                                            "Processing Broker FLUSH(%s.%s)\n",
					    message.db, message.table);
					if (*message.table)
                                        	cacheInvalidateTable(server,
							message.db,
							message.table);
					else
                                        	cacheInvalidateDatabase(
							server,
							message.db);
                                        break;
                                
                                case CMD_CLIENT_DB:
                                        msqlDebug0(MOD_BROKER,
                                               "Processing Broker CLIENT_DB\n");
					if(conArray[message.client].db)
					    free(conArray[message.client].db);
                                        conArray[message.client].db =
                                                strdup(message.db);
                                        conArray[message.client].access =
                                                message.access;
                                        break;

                                case CMD_CLIENT_CLOSE:
                                        count = message.client;
                                        comSock = conArray[count].sock;
                                        msqlDebug2(MOD_BROKER,
                                         "Processing Broker CLOSE (%d on %d)\n",
                                                count, comSock);
					if (comSock < 0)
					{
                                        	msqlDebug2(MOD_BROKER,
	                                         "Close ignored (%d on %d)\n",
                                                count, comSock);
						break;
					}
                                        FD_CLR(comSock,&clientFDs);
                                        shutdown(comSock,2);
                                        close(comSock);
                                        conArray[count].sock = -1;
                                        if(conArray[count].user)
					{
                                        	free(conArray[count].user);
                                        	conArray[count].user = NULL;
					}
                                        if(conArray[count].host)
					{
                                        	free(conArray[count].host);
                                        	conArray[count].host = NULL;
					}
                                        if(conArray[count].db)
					{
                                        	free(conArray[count].db);
                                        	conArray[count].db = NULL;
					}
					numCons--;
                                        break;
				}
			}
                        continue;
                }

                /*
                ** Must be coming from an active client
                */
                count = 0;
		clientsProcessed = 0;
                while(count <= 255)
                {
                        if (conArray[count].sock == -1)
                        {
                                count++;
                                continue;
                        }
                        if (!FD_ISSET(conArray[count].sock,&readFDs))
                        {
                                count++;
                                continue;
                        }

                        /* Got one ! */
			clientsProcessed++;
                        comSock = conArray[count].sock;
			if (netReadPacket(comSock) <= 0)
			{
				msqlDebug1(MOD_GENERAL,
					"Command read on sock %d failed!\n",
					comSock);
				clientCommand = QUIT;
			}
			else
			{
				clientCommand = atoi(packet);
			}
			msqlDebug3(MOD_GENERAL,"Command on sock %d = %d (%s)\n",
				comSock, clientCommand, 
				comTable[clientCommand]);

			/*
			** with it.
			if (clientCommand != QUIT && clientCommand != INIT_DB)
				lockReleaseIpcLock(server);
			*/

			switch(clientCommand)
			{
                            case QUIT:
                                msqlDebug0(MOD_GENERAL,"DB QUIT!\n");
                                FD_CLR(comSock,&clientFDs);
                                shutdown(comSock,2);
                                close(comSock);
                                if(conArray[count].user)
				{
                                	free(conArray[count].user);
                                	conArray[count].user = NULL;
				}
                                if(conArray[count].host)
				{
                                	free(conArray[count].host);
                                	conArray[count].host = NULL;
				}
                                if(conArray[count].db)
				{
                                	free(conArray[count].db);
                                	conArray[count].db = NULL;
				}
                                conArray[count].sock = -1;

                                /* Send broker update request */
				_zeroMessageStruct(&message);
                                message.command = CMD_CLIENT_CLOSE;
                                message.client = count;
                                brokerChildSendMessage(&message);
                                break;


			    case INIT_DB:
				cp=(char *)strtok(packet+2,"\n\r");
				if (!cp)
				{
					netError(comSock,NO_DB_ERROR);
					break;
				}
				strcpy(dbname,cp);
				msqlDebug1(MOD_GENERAL,"DBName = %s\n", dbname);
				conArray[count].access = aclCheckAccess(
					dbname, conArray + count);
				if(conArray[count].access == NO_ACCESS)
				{
					netError(comSock,ACCESS_DENIED_ERROR);
					break;
				}
				if(conArray[count].db)
				{
					free(conArray[count].db);
					conArray[count].db = NULL;
				}
				conArray[count].db = (char *)strdup(dbname);
				if (utilCheckDB(server, dbname) < 0)
				{
					netError(comSock,errMsg);
					break;
				}

                                /* Send broker update */
				_zeroMessageStruct(&message);
                                message.command = CMD_CLIENT_DB;
                                message.access = conArray[count].access;
                                message.client = count;
                                strcpy(message.db,dbname);
                                brokerChildSendMessage(&message);
                                netOK(comSock);
                                break;

			    case QUERY:
				if (!conArray[count].db)
				{
					netError(comSock,NO_DB_ERROR);
					break;
				}
				curSock = comSock;
				cp=(char *)(packet+2);
				arg = (char *)strdup(cp);
				tmpChar = 0;
				if (strlen(arg) > 4096)
				{
					tmpChar = *(arg+4096);
					*(arg+4096) = 0;
				}
				msqlDebug1(MOD_QUERY,"Query = %s",arg);
				if (tmpChar)
					*(arg+4096) = tmpChar;
				aclSetPerms(conArray[count].access);
				query = parseQuery(server, arg,comSock,
					conArray[count].user, 
					conArray[count].db);
				if (query)
				{
					if (haveLock)
					{
                        			lockReleaseIpcLock(server);
						haveLock = 0;
					}
					processQuery(server,query,count,arg);
					parseCleanQuery(query);
				}
				if(arg)
					free(arg);
				/* cacheSyncCache(); */
				break;


			    case EXPLAIN:
				if (!conArray[count].db)
				{
					netError(comSock,NO_DB_ERROR);
					break;
				}
				curSock = comSock;
				cp=(char *)(packet+2);
				while(*cp != ':') cp++;
				arg = (char *)strdup(cp + 1);
				tmpChar = 0;
				if (strlen(arg) > 4096)
				{
					tmpChar = *(arg+4096);
					*(arg+4096) = 0;
				}
				msqlDebug1(MOD_QUERY,"Explain = %s",arg);
				if (tmpChar)
					*(arg+4096) = tmpChar;
				aclSetPerms(conArray[count].access);
				query = parseQuery(server, arg,comSock,
					conArray[count].user, 
					conArray[count].db);
				if (query)
				{
					query->explainOnly = 1;
					processQuery(server,query,count,arg);
					parseCleanQuery(query);
				}
				if(arg)
					free(arg);
				/* cacheSyncCache(); */
				break;


			    case DB_LIST:
				curSock = comSock;
				processListDBs(server, comSock);
				break;

			    case SEQ_INFO:
				if (!conArray[count].db)
				{
					netError(comSock,NO_DB_ERROR);
					break;
				}
				cp = (char *)index(packet,':');
				cp=(char *)strtok(cp+1,
					"\n\r");
				arg = (char *)strdup(cp);
				curSock = comSock;
				processSequenceInfo(server, comSock,
					arg,conArray[count].db);
				if(arg)
					free(arg);
				break;

			    case TABLE_LIST:
				if (!conArray[count].db)
				{
					netError(comSock, NO_DB_ERROR);
					break;
				}
				curSock = comSock;
				processListTables(server, comSock,
					conArray[count].db);
				break;

			    case FIELD_LIST:
				if (!conArray[count].db)
				{
					netError(comSock,NO_DB_ERROR);
					break;
				}
				cp=(char *)strtok(packet+2,
					"\n\r");
				arg = (char *)strdup(cp);
				curSock = comSock;
				processListFields(server, comSock, arg,
					conArray[count].db);
				if(arg)
					free(arg);
				break;

			    case INDEX_LIST:
				if (!conArray[count].db)
				{
					netError(comSock,NO_DB_ERROR);
					break;
				}
				cp=(char *)strtok(packet+2,
					":\n\r");
				arg = (char *)strdup(cp);
				cp=(char *)strtok(NULL,"\n\r");
				arg2 = (char *)strdup(cp);
				curSock = comSock;
				processListIndex(server, comSock, arg2, arg,
					conArray[count].db);
				if(arg)
					free(arg);
				if(arg2)
					free(arg2);
				break;

                            case EXPORT:
                                if (!conArray[count].db)
                                {
                                        netError(comSock,NO_DB_ERROR);
                                        break;
                                }
				arg = arg2 = arg3 = NULL;
                                cp=(char *)strtok(packet+2, ":\n\r");
                                arg = (char *)strdup(cp);
                                cp=(char *)strtok(NULL,":\n\r");
                                arg2 = (char *)strdup(cp);
				cp=(char *)strtok(NULL,":\n\r");
				if (cp)
				{
					arg3 = (char *)strdup(cp);
				}

                                curSock = comSock;
                        	processExportTable(server, comSock,
					server->conArray[count].db, arg, 
					arg2, arg3, NULL);
                                if(arg)
					free(arg);
                                if(arg2)
					free(arg2);
                                if(arg3)
					free(arg3);
                                break;

                            case IMPORT:
                                if (!conArray[count].db)
                                {
                                        netError(comSock,NO_DB_ERROR);
                                        break;
                                }
                                cp=(char *)strtok(packet+2,
                                        ":\n\r");
                                arg = (char *)strdup(cp);
                                cp=(char *)strtok(NULL,":\n\r");
                                arg2 = (char *)strdup(cp);
                                curSock = comSock;
				dummyQuery.curUser = conArray[count].user;
				dummyQuery.curDB = conArray[count].db;
                                processImportTable(server,comSock, 
					conArray[count].db, arg,arg2, 
					&dummyQuery);
                                if(arg)
					free(arg);
                                if(arg2)
					free(arg2);
                                break;

			    case CREATE_DB:
				if (!aclCheckLocal(conArray + count))
				{
					netError(comSock,PERM_DENIED_ERROR);
					break;
				}
				cp=(char *)strtok(packet+2,
					"\n\r");
				arg = (char *)strdup(cp);
				processCreateDB(server, comSock,arg);
				if(arg)
					free(arg);
				break;

			    case COPY_DB:
				if (!aclCheckLocal(conArray + count))
				{
					netError(comSock,PERM_DENIED_ERROR);
					break;
				}
				cp=(char *)strtok(packet+2, ":\n\r");
				arg = (char *)strdup(cp);
				cp=(char *)strtok(NULL, "\n\r");
				arg2 = (char *)strdup(cp);
				processCopyDB(server, comSock,arg, arg2);
				if(arg)
					free(arg);
				if(arg2)
					free(arg2);
				break;

			    case MOVE_DB:
				if (!aclCheckLocal(conArray + count))
				{
					netError(comSock,PERM_DENIED_ERROR);
					break;
				}
				cp=(char *)strtok(packet+2, ":\n\r");
				arg = (char *)strdup(cp);
				cp=(char *)strtok(NULL, "\n\r");
				arg2 = (char *)strdup(cp);
				processMoveDB(server, comSock,arg, arg2);
				if(arg)
					free(arg);
				if(arg2)
					free(arg2);
                                
                                /* Send broker update request */
				_zeroMessageStruct(&message);
                                message.command = CMD_FLUSH_CACHE;
                                message.client = 0;
				strcpy(message.db, arg);
                                brokerChildSendMessage(&message);
				break;

			    case DROP_DB:	
				if (!aclCheckLocal(conArray + count))
				{
					netError(comSock,PERM_DENIED_ERROR);
					break;
				}
				cp=(char *)strtok(packet+2,
					"\n\r");
				arg = (char *)strdup(cp);
				processDropDB(server, comSock,arg);
                                
                                /* Send broker update request */
				_zeroMessageStruct(&message);
                                message.command = CMD_FLUSH_CACHE;
                                message.client = 0;
				strcpy(message.db, arg);
                                brokerChildSendMessage(&message);
				if(arg)
					free(arg);
				break;

			    case RELOAD_ACL:
				if (!aclCheckLocal(conArray + count))
				{
					netError(comSock,PERM_DENIED_ERROR);
					break;
				}
				aclReloadFile(comSock);
				netEndOfList(comSock);
				break;

			    case SHUTDOWN:
				if (!aclCheckLocal(conArray + count))
				{
					netError(comSock,PERM_DENIED_ERROR);
					break;
				}
				netEndOfList(comSock);
				puntServer(-1);
				exit(0);
				break;

			    case SERVER_STATS:
				if (!aclCheckLocal(conArray + count))
				{
					netError(comSock,PERM_DENIED_ERROR);
					break;
				}
				sendServerStats(comSock);
				netEndOfList(comSock);
				break;

			    default:
				netError(comSock, UNKNOWN_COM_ERROR);
				break;
			}
			
                        /*
                        ** Fall out of the loop so that we process
                        ** requests in sequence.  Otherwise we may
                        ** have given away the select lock when we
                        ** actually need it (like INIT_DB)
                        */
                        break;
		}
		if(haveLock)
		{
                       	lockReleaseIpcLock(server);
		}
		if (clientsProcessed == 0)
		{
			/*
			** Something strange is going on.  We fell out
			** of select but didn't do anything.  The client
			** fdset and the connection table must not match
			*/
			printf("Bad client FD_SET in %d!\n", (int)getpid());
			abort();
		}
	}
}


void debugTrap()
{
	/* We use this as a debugger breakpoint */
}


int main(argc,argv)
	int	argc;
	char	*argv[];
{
	msqld	server;
	fd_set	readFDs,
		children;
	int	sock,
		newSock,
		maxSock,
		maxChildSock,
		opt,
		error,
		errFlag = 0,
		lockFD,
		c;
	char	*uname,
		*ipAddr,
		path[MSQL_PATH_LEN],
		pidBuf[10];
	int	pidFD;
        extern  char *optarg;
#ifdef HAVE_SETRLIMIT
	struct	rlimit limit;
#endif
	struct	passwd *pwd;
	struct	stat sbuf;
	struct	sockaddr_in remote;

#ifdef	DEBUG_BSD_MALLOC
	extern char *malloc_options;
	malloc_options = "AJ";
#endif

	serverStartTime = time(NULL);
        printf("\n\nMini SQL Version %s\n",SERVER_VERSION);
        printf("Copyright (c) 1993-94 David J. Hughes\n");
        printf("Copyright (c) 1995-2012 Hughes Technologies Pty Ltd.\n");
        printf("All rights reserved.\n\n");

        /* Try to set the locale but don't bail if it fails (freeBSD) */
        setlocale(LC_ALL, "");

	/*
	** Handle the command line args
	*/
	*confFile = 0;
        while((c=getopt(argc,argv,"f:"))!= -1)
        {
                switch(c)
                {
                        case 'f':
                                if (*confFile)
                                        errFlag++;
                                else
                                        strcpy(confFile,optarg);
                                break;
                        case '?':
                                errFlag++;
                                break;
                }
	}
	if (errFlag)
	{
		usage();
		exit(1);
	}

	if (*confFile == 0)
	{
		snprintf(confFile,sizeof(confFile),"%s/msql.conf", INST_DIR);
	}
	printf("\tLoading configuration from '%s'.\n",confFile);
	configLoadFile(confFile);

	bzero(&server, sizeof(server));
        globalServer = &server;
	server.config.instDir=strdup(configGetCharEntry("general","inst_dir"));
	server.config.dbDir = strdup(configGetCharEntry("general", "db_dir"));
	server.config.readOnly = configGetIntEntry("system", "read_only");
	server.config.msyncTimer = configGetIntEntry("system", "msync_timer");
        server.config.sortMaxMem = configGetIntEntry("system","sort_max_mem");
        server.config.tableCache = configGetIntEntry("system","table_cache");
        server.config.cacheDescriptors = server.config.tableCache * 
		(NUM_INDEX + 2);
        server.config.needFileLock = 1;
        server.config.hasBroker = 1;


	/*
	** Check out the file descriptor limit
	**
	** Some boxes have broken getrlimit() implementations (or
	** missing RLIMIT_NOFILE) so we try to use sysconf first
	**
	** Watch out for broken BSDI here.  It looks like BSDI returns
	** the kernel MAX_FILES not the per process MAX_FILE in
	** getrlimit() which kinda sucks if you want to use select()!
	*/

#	ifdef HAVE_RLIMIT_NOFILE
		getrlimit(RLIMIT_NOFILE,&limit);
		limit.rlim_cur = (limit.rlim_max > 256)? 256 : limit.rlim_max;
		setrlimit(RLIMIT_NOFILE,&limit);
		getrlimit(RLIMIT_NOFILE,&limit);
		maxCons = limit.rlim_cur-globalServer->config.cacheDescriptors;
		printf("\tServer process reconfigured to accept %d ",maxCons);
		printf("connections.\n");
#	else
#	ifdef HAVE_SYSCONF
		maxCons = sysconf(_SC_OPEN_MAX)-
			globalServer->config.cacheDescriptors;
		printf("\tServer process reconfigured to accept %d ",maxCons);
		printf("connections.\n");
#  	else
		maxCons = 64 - globalServer->config.cacheDescriptors;
		printf("\tDynamic server config failed!\n");
		printf("\tServer can accept %d connections.\n", maxCons);
#	endif
#	endif

	if (maxCons > 254)
		maxCons = 254;

	/*
	** We have to do this now.  If msql is configured to run as nobody
	** then nobody would have to have write perms for the INST_DIR
	** This way we can start as root and then run as the MSQL_USER when
	** we're setup properly
	*/
	initDebug(1);
	netInitialise();
	initServer();

	(void)snprintf(path, MSQL_PATH_LEN, "%s", 
		(char *)configGetCharEntry("general","pid_file"));
	pidFD = open(path, O_CREAT|O_TRUNC|O_RDWR, 0700);
	if (pidFD < 0)
	{
		perror("Error : Couldn't open PID file");
		exit(1);
	}
	else
	{
		if (lockNonBlockingLock(pidFD) < 0)
		{
			printf("\nError : Cannot lock PID file!\n\n");
			exit(1);
		}
		sprintf(pidBuf,"%d", (int)getpid());
		write(pidFD, pidBuf, strlen(pidBuf) + 1);
	}


	/*
	** Create the select lock file 
	*/
	sprintf(path,"%s/msql.lock",INST_DIR);
	lockFD = open(path,O_CREAT|O_TRUNC|O_RDWR,0700);
	close(lockFD);

	/*
	** Are we running as the right user?
	*/
	pwd = getpwuid(getuid());
	if (!pwd)
	{
		printf("\nError!  No username for our UID (%d)\n\n",
			(int)getuid());
		exit(1);
	}
	uname = (char *)configGetCharEntry("general","msql_user");
	if (strcmp(uname,pwd->pw_name) != 0)
	{
		pwd = getpwnam(uname);
		if (!pwd)
		{
			printf("\nError!  Unknown username (%s)\n\n", uname);
			exit(1);
		}
		chown(path,pwd->pw_uid, pwd->pw_gid);
		chown("/proc/self/fd",pwd->pw_uid,pwd->pw_gid);
		if (setuid(pwd->pw_uid) < 0)
		{
			printf("\nError!  Can't run as user '%s'\n\n",uname);
			exit(1);
		}
	}
	printf("\tServer running as user '%s'.\n",uname);
	printf("\tServer mode is %s.\n",
		configGetIntEntry("system","read_only")?"Read-Only":"Read/Write");

	/*
	** Ensure that the correct user owns the database files
	*/
	snprintf(path,MSQL_PATH_LEN, "%s/msqldb",
		(char *)configGetCharEntry("general","inst_dir"));
	if (stat(path,&sbuf) < 0)
	{
		printf("\nError!  Can't stat '%s'\n\n",path);
		exit(1);
	}

	if (sbuf.st_uid != getuid())
	{
		printf("\nError!  '%s' is not owned by '%s'\n\n",path,
			pwd->pw_name);
		exit(1);
	}

	/*
	** OK, on with the show
	*/

	umask(0);
	numCons=0;
	maxSock = 0;
	/* chdir(msqlHomeDir); */
	tableCleanTmpDir(NULL);

#	ifdef BETA_RELEASE
	timeRemain = (863060000 + 60*24*60*60) - time(NULL);
	if (timeRemain < 0)
	{
		printf("\nThis Beta test release has expired!\n");
		printf("Check out www.Hughes.com.au for a new release!\n\n");
		exit(1);
	}
	else
	{
		printf("\n\tThis Beta test release expires in %d days.\n\n",
			(int)(timeRemain / (24*60*60)));
	}
#	endif

	chmod(path,0700);
	umask(0077);
	setupSignals();
	(void)bzero(&children,sizeof(fd_set));
	(void)bzero(conArray,sizeof(conArray));
	maxChildSock = brokerStartChildren(&server,&children, argc,argv);
	strcpy(PROGNAME,"msqld");
	aclLoadFile(1);
	msqlDebug0(MOD_GENERAL,
		"miniSQL debug mode.  Waiting for connections.\n");

	while(1)
	{
		/*
		** Setup the read fd set
		*/
		FD_ZERO(&readFDs);
		if (IPsock >= 0)
			FD_SET(IPsock,&readFDs);
		if (UNIXsock >= 0)
			FD_SET(UNIXsock,&readFDs);
		brokerAddChildSockets(&readFDs);


		/* Wait for some action */
		if(select((maxCons+globalServer->config.cacheDescriptors),
			&readFDs,0,0,0) < 0)
		{
			continue;
		}

		/* 
		** Is there a child notifying us? 
		*/

		brokerCheckChildren(&readFDs);

		/*
		** How about a new connection?  Grab the lock so we
		** can ensure the broker notification gets out before
		** the new connection becomes active
		*/

		sock = 0;
		if (IPsock >= 0)
		{
			if (FD_ISSET(IPsock,&readFDs))
			{
				sock = IPsock;
			}
		}
		if (UNIXsock >= 0)
		{
			if (FD_ISSET(UNIXsock,&readFDs))
			{
				sock = UNIXsock;
			}
		}
		if (sock)
		{
			struct	sockaddr_in 	cAddr;
			struct	sockaddr	dummy;
			int	cAddrLen,
				dummyLen;

			bzero(&cAddr, sizeof(cAddr));
			cAddrLen = sizeof(struct sockaddr_in);
			newSock = accept(sock, (struct sockaddr *)&cAddr, 
				&cAddrLen);
			if(newSock < 0)
			{
				perror("Error in accept ");
				puntServer(-1);
				exit(1);
			}
			if (newSock > maxSock)
				maxSock = newSock;
			dummyLen = sizeof(struct sockaddr);
			if (getsockname(newSock,&dummy, &dummyLen) < 0)
			{
				perror("Error on new connection socket");
				continue;
			}
			if (conArray[newSock].db)
			{
				if(conArray[newSock].db)
				{
					free(conArray[newSock].db);
					conArray[newSock].db = NULL;
				}
				if(conArray[newSock].host)
				{
					free(conArray[newSock].host);
					conArray[newSock].host = NULL;
				}
				if(conArray[newSock].user)
				{
					free(conArray[newSock].user);
					conArray[newSock].user = NULL;
				}
			}

			/*
			** Are we over the connection limit
			*/
			numCons++;
			if (numCons > maxCons)
			{
				numCons--;
				netError(newSock,CON_COUNT_ERROR);
				shutdown(newSock,2);
				close(newSock);
				continue;
			}


			/*
			** store the connection details
			*/

			msqlDebug1(MOD_GENERAL,
				"New connection received on %d\n", newSock);
			error = 0;
			if (sock == IPsock)
			{
				int	addrLen;
				struct	hostent *hp;

				addrLen = sizeof(struct sockaddr);
				getpeername(newSock, (struct sockaddr *)
					&remote, &addrLen);
				ipAddr = inet_ntoa(remote.sin_addr);
				if (ipAddr)
				{
					conArray[newSock].clientIP =
						strdup(ipAddr);
				}
				if (configGetIntEntry("system","host_lookup") == 1)
				{
					/* 
					** Validate remote host 
					*/

     					if(auth_hostname(&remote)) 
					{
						netError(newSock,
							BAD_HOST_ERROR);
						error = 1;
						numCons--;
						shutdown(newSock,2);
						close(newSock);
					}

					/*
					** Grab the hostname for later
					*/
					hp = (struct hostent *)gethostbyaddr(
					        (char *)&remote.sin_addr,
						sizeof(remote.sin_addr),
						AF_INET);
				}
				else
				{
					hp = NULL;
				}

				if (!hp)
				{
					conArray[newSock].host = (char *)
						strdup("UNKNOWN_HOST");
					msqlDebug0(MOD_GENERAL,
						"Host=UNKNOWN_HOST\n");
				}
				else
				{
					conArray[newSock].host = (char *)
						strdup((char *)hp->h_name);
					msqlDebug1(MOD_GENERAL,"Host = %s\n",
						hp->h_name);
				}
			}
			else
			{
				conArray[newSock].host = NULL;
				conArray[newSock].clientIP = NULL;
				msqlDebug0(MOD_GENERAL,"Host = UNIX domain\n");
			}

			conArray[newSock].connectTime = time(NULL);
			conArray[newSock].numQueries = 0;

			if (!error)
			{
				opt=1;
				setsockopt(newSock,SOL_SOCKET,SO_KEEPALIVE,
					(char *) &opt, sizeof(opt));
                                /*fcntl(newSock, F_SETFL, O_NONBLOCK);*/
				snprintf(packet, PKT_LEN,
					"0:%d:%s\n",
					PROTOCOL_VERSION,SERVER_VERSION);
				netWritePacket(newSock);
				if (netReadPacket(newSock) <=0)
				{
					netError(newSock,HANDSHAKE_ERROR);
					numCons--;
					shutdown(newSock,2);
					close(newSock);
					conArray[newSock].host = NULL;
					conArray[newSock].clientIP = NULL;
				}
				else
				{
					uname = (char *)strtok(packet,"\n");
					msqlDebug1(MOD_GENERAL,
						"User = %s\n",uname);
					if(conArray[newSock].user)
						free(conArray[newSock].user);
					conArray[newSock].user = (char *)
						strdup(uname);
					brokerNotifyAllChildren(CMD_CLIENT_OPEN,
						-1, newSock, uname, NULL, NULL,
						0, conArray[newSock].clientIP);
					netEndOfList(newSock);
				}
			}
			continue;
		}
	}
}
